"""
A hidden command, used by support to assist with migrating documents
that need to be updated
"""

from pathlib import Path
from typing import Optional

import click
import yaml

import spyctl.config.configs as cfg
import spyctl.resources.policies as p
import spyctl.spyctl_lib as lib
from spyctl import api
from spyctl.commands import apply

# ----------------------------------------------------------------- #
#                    Hidden Update Subcommand                       #
# ----------------------------------------------------------------- #


@click.group(
    "update", cls=lib.CustomSubGroup, hidden=True, epilog=lib.SUB_EPILOG
)
@click.help_option("-h", "--help", hidden=True)
def update():
    pass


@update.command("response-actions", cls=lib.CustomCommand)
@click.help_option("-h", "--help", hidden=True)
@click.option(
    "-b",
    "--backup-file",
    "backup_file",
    help="location to place policy backups",
    type=click.Path(exists=True, writable=True, file_okay=False),
)
def update_response_actions(backup_file=None):
    handle_update_response_actions(backup_file)


@update.command("policy-modes", cls=lib.CustomCommand)
@click.help_option("-h", "--help", hidden=True)
@click.option(
    "-b",
    "--backup-file",
    "backup_file",
    help="location to place policy backups",
    required=True,
    type=click.Path(exists=True, writable=True, file_okay=False),
)
def update_policy_modes(backup_file=None):
    handle_update_policy_modes(backup_file)


# ----------------------------------------------------------------- #
#                         Update Handlers                           #
# ----------------------------------------------------------------- #


def handle_update_response_actions(backup_dir: Optional[str]):
    ctx = cfg.get_current_context()
    policies = api.get_policies(*ctx.get_api_data())
    if backup_dir:
        backup_path = Path(backup_dir)

    for policy in policies:
        pol_data = p.policies_output([policy])
        if backup_path:
            try:
                uid = pol_data[lib.METADATA_FIELD][lib.METADATA_UID_FIELD]
                outfile = Path.joinpath(backup_path, uid)
                print(f"Backing up {uid} tp {str(outfile)}")
                outfile.write_text(yaml.dump(pol_data, sort_keys=False))
            except Exception:
                import traceback

                s = traceback.format_exc()
                print(s)
                lib.err_exit("Error saving policy backups.. canceling")
        del pol_data[lib.SPEC_FIELD][lib.RESPONSE_FIELD]
        pol_data = p.Policy(pol_data).as_dict()
        apply.handle_apply_policy(pol_data)


def handle_update_policy_modes(backup_dir: Optional[str]):
    ctx = cfg.get_current_context()
    print("Updating Guardian Policies")
    policies = api.get_policies(*ctx.get_api_data())
    if backup_dir:
        backup_path = Path(backup_dir)
    for policy in policies:
        pol_data = p.policies_output([policy])
        if backup_path:
            try:
                uid = pol_data[lib.METADATA_FIELD][lib.METADATA_UID_FIELD]
                outfile = Path.joinpath(backup_path, uid)
                print(f"Backing up {uid} tp {str(outfile)}")
                outfile.write_text(yaml.dump(pol_data, sort_keys=False))
            except Exception:
                import traceback

                s = traceback.format_exc()
                print(s)
                lib.err_exit("Error saving policy backups.. canceling")
        mode = pol_data[lib.SPEC_FIELD].get(
            lib.POL_MODE_FIELD, lib.POL_MODE_ENFORCE
        )
        if mode not in lib.POL_MODES:
            mode = lib.POL_MODE_ENFORCE
        lib.try_log(
            f"Setting mode for policy"
            f" '{policy[lib.METADATA_FIELD][lib.METADATA_UID_FIELD]}' to"
            f" '{mode}'"
        )
        pol_data[lib.SPEC_FIELD][lib.POL_MODE_FIELD] = mode
        pol_data = p.Policy(pol_data).as_dict()
        apply.handle_apply_policy(pol_data)
    print("Updating Suppression Policies")
    policies = api.get_policies(
        *ctx.get_api_data(),
        params={lib.METADATA_TYPE_FIELD: lib.POL_TYPE_TRACE},
    )
    for policy in policies:
        if backup_path:
            try:
                uid = policy[lib.METADATA_FIELD][lib.METADATA_UID_FIELD]
                outfile = Path.joinpath(backup_path, uid)
                print(f"Backing up {uid} tp {str(outfile)}")
                outfile.write_text(yaml.dump(policy, sort_keys=False))
            except Exception:
                import traceback

                s = traceback.format_exc()
                print(s)
                lib.err_exit("Error saving policy backups.. canceling")
        mode = policy[lib.SPEC_FIELD].get(
            lib.POL_MODE_FIELD, lib.POL_MODE_ENFORCE
        )
        if mode not in lib.POL_MODES:
            mode = lib.POL_MODE_ENFORCE
        lib.try_log(
            f"Setting mode for suppression policy"
            f" '{policy[lib.METADATA_FIELD][lib.METADATA_UID_FIELD]}' to"
            f" '{mode}'"
        )
        policy[lib.SPEC_FIELD][lib.POL_MODE_FIELD] = mode
        apply.handle_apply_policy(policy)



"""Handles the config subcommand group for spyctl."""

import click

import spyctl.spyctl_lib as lib
from spyctl import cli

import spyctl.config as _c

# ----------------------------------------------------------------- #
#                         Config Subcommand                         #
# ----------------------------------------------------------------- #


@click.group("config", cls=lib.CustomSubGroup, epilog=lib.SUB_EPILOG)
@click.help_option("-h", "--help", hidden=True)
def config():
    """Modify spyctl config files."""


@config.command("delete-context", cls=lib.CustomCommand, epilog=lib.SUB_EPILOG)
@click.help_option("-h", "--help", hidden=True)
@click.option(
    "-g",
    "--global",
    "force_global",
    is_flag=True,
    help="When operating within a spyctl workspace, this deletes a context"
    " from the global spyctl configuration file.",
)
@click.option(
    "-y",
    "--yes",
    "--assume-yes",
    is_flag=True,
    help='Automatic yes to prompts; assume "yes" as answer to all prompts and'
    " run non-interactively.",
)
@click.argument(
    "name",
    type=_c.configs.ContextsParam(),
)
def delete_context(name, force_global, yes=False):
    """Delete the specified context from a spyctl configuration file.

    NAME is the name of the context to delete."""
    if yes:
        cli.set_yes_option()
    _c.configs.delete_context(name, force_global)


@config.command(
    "delete-apisecret", cls=lib.CustomCommand, epilog=lib.SUB_EPILOG
)
@click.help_option("-h", "--help", hidden=True)
@click.option(
    "-y",
    "--yes",
    "--assume-yes",
    is_flag=True,
    help='Automatic yes to prompts; assume "yes" as answer to all prompts and'
    " run non-interactively.",
)
@click.argument("name", type=_c.secrets.SecretsParam())
def delete_apisecret(name, yes=False):
    """Delete the specified apisecret from a spyctl configuration file.

    NAME is the name of the apisecret to delete."""
    if yes:
        cli.set_yes_option()
    _c.secrets.delete_secret(name)


@config.command(
    "current-context", cls=lib.CustomCommand, epilog=lib.SUB_EPILOG
)
@click.help_option("-h", "--help", hidden=True)
@click.option(
    "-g",
    "--global",
    "force_global",
    is_flag=True,
    help="When operating within a spyctl workspace, this displays the"
    " current context in the global configuration.",
)
def current_context(force_global):
    """Display the current-context."""
    _c.configs.current_context(force_global)


@config.command("get-contexts", cls=lib.CustomCommand, epilog=lib.SUB_EPILOG)
@click.help_option("-h", "--help", hidden=True)
@click.argument("name", required=False, type=_c.configs.ContextsParam())
@click.option(
    "-o",
    "--output",
    default=lib.OUTPUT_DEFAULT,
    type=click.Choice(
        [lib.OUTPUT_DEFAULT, lib.OUTPUT_WIDE], case_sensitive=False
    ),
)
@click.option(
    "-g",
    "--global",
    "force_global",
    is_flag=True,
    help="When operating within a spyctl workspace, this displays only"
    " contexts within the global spyctl configuration file.",
)
@click.option(
    "-w",
    "--workspace",
    "force_workspace",
    is_flag=True,
    help="When operating within a spyctl workspace, this displays only"
    " contexts within the workspace configuration file.",
)
def get_contexts(force_global, force_workspace, output, name=None):
    """Describe one or many contexts.

    NAME is the name of a specific context to view.

    The default behavior is to show all of the contexts accessible to
    the current working directory. If not using a workspace, this is
    only the contexts in the global config. See --help for \"spyctl
    config init\" for more details.
    """
    if force_global and force_workspace:
        cli.try_log(
            "Both global and workspace flags set; defaulting to global"
        )
    _c.configs.get_contexts(name, force_global, force_workspace, output)


@config.command("get-apisecrets", cls=lib.CustomCommand, epilog=lib.SUB_EPILOG)
@click.help_option("-h", "--help", hidden=True)
@click.argument("name", required=False, type=_c.secrets.SecretsParam())
@click.option(
    "-o",
    "--output",
    default=lib.OUTPUT_DEFAULT,
    type=click.Choice(
        lib.OUTPUT_CHOICES + [lib.OUTPUT_WIDE], case_sensitive=False
    ),
)
def get_api_secrets(output, name=None):
    """Describe one or many apisecrets."""
    _c.secrets.handle_get_secrets(name, output)


@config.command("init-workspace", cls=lib.CustomCommand, epilog=lib.SUB_EPILOG)
@click.option(
    "-y",
    "--yes",
    "--assume-yes",
    is_flag=True,
    help='Automatic yes to prompts; assume "yes" as answer to all prompts and'
    " run non-interactively.",
)
@click.help_option("-h", "--help", hidden=True)
def init_workspace(yes=False):
    """Initialize a workspace.

    This command creates a '.spyctl/config' file in the current working
    directory. Workspaces are a spyctl configuration file local to a
    specific directory. They allow you to create contexts only
    accessible from the directory subtree from where the config file
    resides. They also allow you to set a current context for a
    directory subtree.

    This is helpful if you're working on a specific service or
    container and want spyctl to return data relevant only to that
    application.

    For example:
    If your cwd is ``$HOME/myproject/`` and you issue the command
    ``spyctl config current-context`` you will be shown the current
    context in the global configuration. But if create initialize
    a workspace and create a context, you will notice that your
    current context is the one set within the workspace configuration
    file.

    For example:

      # Create a workspace in the current working directory.\n
      spyctl config init-workspace

      # Create a context specific to a Linux Service.\n
      spyctl config set-context --cgroup systemd:/system.slice/my_app.service
      --org my_organization --secret my_secret my_app_context

      # Show the current context and see my_app_context in the output.\n
      spyctl config current-context

    Executing spyctl outside of a workspace directory or any of its
    subdirectories will revert the tool to using the current context in
    the global configuration file.
    """
    if yes:
        cli.set_yes_option()
    _c.configs.init()


@config.command("set-context", cls=lib.CustomCommand, epilog=lib.SUB_EPILOG)
@click.help_option("-h", "--help", hidden=True)
@click.option(
    "-g",
    "--global",
    "force_global",
    is_flag=True,
    help="When operating within a spyctl workspace, this sets a context"
    " in the global spyctl configuration file.",
)
@click.option(
    "-u",
    "--use-context",
    "use_ctx",
    is_flag=True,
    help="Use this context when set. Sets the current-context field in the"
    " config file. The first context added to a config is automatically set as"
    " current-context.",
)
@click.option(
    "-s",
    "--secret",
    help="Name of api config secret.",
    metavar="",
    required=True,
    type=_c.secrets.SecretsParam(),
)
@click.option(
    "-o",
    "--organization",
    "--org",
    help="ID or name of Spyderbat organization.",
    metavar="",
    required=True,
)
@click.option(
    "-c",
    "--cluster",
    help="Name or Spyderbat ID of Kubernetes cluster.",
    metavar="",
    type=lib.ListParam(),
)
@click.option(
    "-n",
    "--namespace",
    help="Name or Spyderbat ID of Kubernetes namespace.",
    metavar="",
    type=lib.ListParam(),
)
@click.option(
    "-p",
    "--pod",
    help="Name or Spyderbat ID of Kubernetes pod.",
    metavar="",
    type=lib.ListParam(),
)
@click.option(
    "-m",
    "--machines",
    help="Name of machine group, or name or Spyderbat ID of a machine (node).",
    metavar="",
    type=lib.ListParam(),
)
@click.option(
    "-i",
    "--image",
    help="Name of container image, wildcards allowed.",
    metavar="",
    type=lib.ListParam(),
)
@click.option(
    "-d",
    "--image-id",
    help="Container image ID.",
    metavar="",
    type=lib.ListParam(),
)
@click.option(
    "-N",
    "--container-name",
    help="Name of specific container.",
    metavar="",
    type=lib.ListParam(),
)
@click.option(
    "-C",
    "--cgroup",
    help="Linux service cgroup.",
    metavar="",
    type=lib.ListParam(),
)
@click.argument("name")
def set_context(name, secret, force_global, use_ctx, **context):
    """Set a context entry in a spyctl configuration file, or update an
    existing one.
    """
    context = {
        key: value for key, value in context.items() if value is not None
    }
    _c.configs.set_context(name, secret, force_global, use_ctx, **context)


@config.command("set-apisecret", cls=lib.CustomCommand, epilog=lib.SUB_EPILOG)
@click.help_option("-h", "--help", hidden=True)
@click.argument("name", required=False)
@click.option(
    "-k",
    "--apikey",
    "--api-key",
    "api_key",
    help="API key generated via the Spyderbat UI",
    metavar="",
)
@click.option(
    "-u",
    "--apiurl",
    "--api-url",
    "api_url",
    help=f"URL target for api queries. Default: {lib.DEFAULT_API_URL}",
    metavar="",
)
def set_apisecrets(api_key=None, api_url=None, name=None):
    """
    Set a new entry in the spyctl secrets file, or update an existing one.
    """
    _c.secrets.set_secret(name, api_url, api_key)


@config.command("use-context", cls=lib.CustomCommand, epilog=lib.SUB_EPILOG)
@click.help_option("-h", "--help", hidden=True)
@click.option(
    "-g",
    "--global",
    "force_global",
    is_flag=True,
    help="When operating within a spyctl workspace, this changes the"
    " current context in the global configuration file.",
)
@click.argument("name", type=_c.configs.ContextsParam())
def use_context(name, force_global):
    """Set the current-context in a spyctl configuration file."""
    _c.configs.use_context(name, force_global)


@config.command("view", cls=lib.CustomCommand, epilog=lib.SUB_EPILOG)
@click.help_option("-h", "--help", hidden=True)
@click.option(
    "-g",
    "--global",
    "force_global",
    is_flag=True,
    help="When operating within a spyctl workspace, this displays"
    " the global spyctl configuration file.",
)
@click.option(
    "-w",
    "--workspace",
    "force_workspace",
    is_flag=True,
    help="When operating within a spyctl workspace, this displays"
    " only the workspace configuration file.",
)
@click.option(
    "-o",
    "--output",
    default=lib.OUTPUT_DEFAULT,
    type=click.Choice(lib.OUTPUT_CHOICES, case_sensitive=False),
)
def view(force_global, force_workspace, output):
    """View the current spyctl configuration file. If operating
    within a workspace the default behavior displays a merged
    configuration including contexts from the global config and any
    other workspace configuration files from cwd to root.
    """
    if force_global and force_workspace:
        cli.try_log(
            "Both global and workspace flags set; defaulting to global"
        )
    _c.configs.view_config(force_global, force_workspace, output)



"""Handles the merge command for spyctl."""

import time
from typing import IO, Dict, List, Optional, Union

import click

import spyctl.config.configs as cfgs
import spyctl.filter_resource as filt
import spyctl.merge_lib as m_lib
import spyctl.resources.baselines as b
import spyctl.resources.deviations as dev
import spyctl.resources.policies as p
import spyctl.resources.resources_lib as r_lib
import spyctl.resources.suppression_policies as sp
import spyctl.schemas_v2 as schemas
import spyctl.spyctl_lib as lib
from spyctl import api, cli
from spyctl.commands import apply

POLICIES = None
FINGERPRINTS = None
DEVIATIONS = None
MATCHING = "matching"
ALL = "all"

YES_EXCEPT = False

# ----------------------------------------------------------------- #
#                         Merge Subcommand                          #
# ----------------------------------------------------------------- #


@click.command("merge", cls=lib.CustomCommand, epilog=lib.SUB_EPILOG)
@click.help_option("-h", "--help", hidden=True)
@click.option(
    "-f",
    "--filename",
    help="Target file(s) of the merge.",
    metavar="",
    type=lib.FileList(),
    cls=lib.MutuallyExclusiveEatAll,
    mutually_exclusive=["policy"],
)
@click.option(
    "-p",
    "--policy",
    is_flag=False,
    flag_value=ALL,
    default=None,
    help="Target policy name(s) or uid(s) of the merge. If supplied with no"
    " argument, set to 'all'.",
    metavar="",
    type=lib.ListParam(),
    cls=lib.MutuallyExclusiveOption,
    mutually_exclusive=["filename"],
)
@click.option(
    "-w",
    "--with-file",
    "with_file",
    help="File to merge into target.",
    metavar="",
    type=click.File(),
    cls=lib.MutuallyExclusiveOption,
    mutually_exclusive=["with_policy"],
)
@click.option(
    "-P",
    "--with-policy",
    "with_policy",
    help="Policy uid to merge with target. If supplied with no argument then"
    " spyctl will attempt to find a policy matching the uid in the"
    " target's metadata.",
    metavar="",
    is_flag=False,
    flag_value=MATCHING,
    cls=lib.MutuallyExclusiveOption,
    mutually_exclusive=["with_file"],
)
@click.option(
    "-l",
    "--latest",
    is_flag=True,
    help=f"Merge file with latest records using the value of"
    f" '{lib.LATEST_TIMESTAMP_FIELD}' in the target's '{lib.METADATA_FIELD}'."
    " This replaces --start-time.",
    metavar="",
)
@click.option(
    "-o",
    "--output",
    default=lib.OUTPUT_DEFAULT,
    type=click.Choice(lib.OUTPUT_CHOICES, case_sensitive=False),
)
@click.option(
    "-t",
    "--start-time",
    "st",
    help="Start time of the query for fingerprints to merge."
    " Only used if --latest, --with-file, and --with-policy are not set."
    " Default is 24 hours ago.",
    default="24h",
    type=lib.time_inp,
)
@click.option(
    "-e",
    "--end-time",
    "et",
    help="End time of the query for fingerprints to merge."
    " Only used if --with-file and --with-policy are not set."
    " Default is now.",
    default=time.time(),
    type=lib.time_inp,
)
@click.option(
    "-O",
    "--output-to-file",
    help="Should output merge to a file. Unique filename created from the name"
    " in the object's metadata.",
    is_flag=True,
)
@click.option(
    "--full-diff",
    is_flag=True,
    help="A diff summary is shown by default, set this flag to show the full"
    " object when viewing a diff following a merge. (All changes to the object"
    " are shown in the summary).",
)
@click.option(
    "-y",
    "--yes",
    "--assume-yes",
    is_flag=True,
    help='Automatic yes to prompts; assume "yes" as answer to all prompts and'
    " run non-interactively.",
    cls=lib.MutuallyExclusiveOption,
    mutually_exclusive=["yes_except"],
)
@click.option(
    "-Y",
    "--yes-except",
    "--assume-yes-except-review",
    is_flag=True,
    help='Automatic yes to merge prompts; assume "yes" as answer to all merge'
    " prompts but still prompts review of policy updates before applying.",
    cls=lib.MutuallyExclusiveOption,
    mutually_exclusive=["yes"],
)
@click.option(
    "--include-network/--exclude-network",
    help="Include or exclude network data in the merge."
    " Default is to include network data in the merge.",
    default=True,
)
@click.option(
    "-a",
    "--api",
    "use_api",
    metavar="",
    default=False,
    hidden=True,
    is_flag=True,
)
@lib.colorization_option
def merge(
    filename,
    policy,
    output,
    st,
    et,
    include_network,
    colorize,
    yes=False,
    yes_except=False,
    with_file=None,
    with_policy=None,
    latest=False,
    output_to_file=False,
    use_api=False,
    full_diff=False,
):
    """Merge target Baselines and Policies with other Resources.

      Merging in Spyctl requires a target Resource (e.g. a Baseline or Policy
    document you are maintaining) and a Resource to merge into the target.
    A target can either be a local file supplied using the -f option or a policy
    you've applied to the Spyderbat Backend supplied with the -p option.
    By default, target's are merged with deviations if they are applied policies,
    otherwise they are merged with relevant* Fingerprints from the last 24
    hours to now. Targets may also be merged with local files with the -w option
    or with data from an existing applied policy using the -P option.

      When merging a single local file with another resource, the output will
    be sent to stdout. WARNING: Do not redirect output to the same file you
    used as input. You may use the -O flag to output the merged data to a
    unique file with a name generate by Spyctl.

      When bulk merging local files, the output for each merge operation will
    be outputted to unique files generated by Spyctl (the same as supplying the
    -O flag mentioned above).

      When merging existing applied policies in bulk or individually, the default
    destination for the output will be to apply it directly to the Spyderbat Backend (you
    will have a chance to review the merge before any changes are applied).
    This removes the requirement to deal with local files when managing policies. However,
    it is a good idea to back up policies in a source-control repository. You can also
    use the -O operation to send the output of this merge to a local file.

    \b
    Examples:
      # merge a local policy file with data from the last
      # 24hrs to now:
      spyctl merge -f policy.yaml\n
    \b
      # merge a local policy file with data from its
      # latestTimestamp field to now:
      spyctl merge -f policy.yaml --latest\n
    \b
      # merge an existing applied policy with data from the
      # last 24hrs to now:
      spyctl merge -p <NAME_OR_UID>\n
    \b
      # Bulk merge all existing policies with data from the
      # last 24hrs to now:
      spyctl merge -p\n
    \b
      # Bulk merge multiple policies with data from the
      # last 24hrs to now:
      spyctl merge -p <NAME_OR_UID1>,<NAME_OR_UID2>\n
    \b
      # Bulk merge all files in cwd matching a pattern with data
      # from the last 24hrs to now:
      spyctl merge -f *.yaml\n
    \b
      # merge an existing applied policy with a local file:
      spyctl merge -p <NAME_OR_UID> --with-file fingerprints.yaml\n
    \b
      # merge a local file with data from an existing applied policy
      spyctl merge -f policy.yaml -P <NAME_OR_UID>\n
    \b
      # merge a local file with a valid UID in its metadata with the matching
      # policy in the Spyderbat Backend
      spyctl merge -f policy.yaml -P

    * Each policy has one or more Selectors in its spec field,
    relevant Fingerprints are those that match those Selectors.

    For time field options such as --start-time and --end-time you can
    use (m) for minutes, (h) for hours (d) for days, and (w) for weeks back
    from now or provide timestamps in epoch format.

    Note: Long time ranges or "get" commands in a context consisting of
    multiple machines can take a long time.
    """  # noqa E501
    if yes or yes_except:
        cli.set_yes_option()
    if not colorize:
        lib.disable_colorization()
    if output == lib.OUTPUT_DEFAULT:
        output = lib.OUTPUT_YAML
    handle_merge(
        filename,
        policy,
        with_file,
        with_policy,
        st,
        et,
        latest,
        output,
        output_to_file,
        yes_except,
        include_network,
        use_api,
        full_diff,
    )


# ----------------------------------------------------------------- #
#                         Merge Handlers                            #
# ----------------------------------------------------------------- #


def handle_merge(
    filename_target: List[IO],
    policy_target: List[str],
    with_file: IO,
    with_policy: str,
    st,
    et,
    latest: bool,
    output: str,
    output_to_file: bool = False,
    yes_except: bool = False,
    merge_network: bool = True,
    do_api=False,
    full_diff=False,
):
    if do_api:
        ctx = cfgs.get_current_context()
        if not filename_target or not with_file:
            cli.err_exit("api test only with local files")
        r_data = lib.load_file_for_api_test(filename_target[0])
        w_data = lib.load_file_for_api_test(with_file)
        merged_data = api.api_merge(*ctx.get_api_data(), r_data, w_data)
        cli.show(merged_data, lib.OUTPUT_RAW)
        return
    global POLICIES, YES_EXCEPT
    YES_EXCEPT = yes_except
    if not POLICIES and (with_policy or policy_target):
        ctx = cfgs.get_current_context()
        POLICIES = api.get_policies(*ctx.get_api_data())
        POLICIES.sort(key=lambda x: x[lib.METADATA_FIELD][lib.NAME_FIELD])
        if not POLICIES and policy_target:
            cli.err_exit("No policies to merge.")
        elif not POLICIES and with_policy:
            cli.err_exit("No policies to merge with.")
    if filename_target:
        if len(filename_target) > 1 or output_to_file:
            output_dest = lib.OUTPUT_DEST_FILE
        else:
            output_dest = lib.OUTPUT_DEFAULT
        filename_target.sort(key=lambda x: x.name)
        for file in filename_target:
            target = load_target_file(file)
            target_name = f"local file '{file.name}'"
            resrc_kind = target.get(lib.KIND_FIELD)
            if resrc_kind not in [lib.BASELINE_KIND, lib.POL_KIND]:
                cli.try_log(
                    f"The 'merge' command is not supported for {resrc_kind}",
                    is_warning=True,
                )
                continue
            with_obj = get_with_obj(
                target,
                target_name,
                with_file,
                with_policy,
                st,
                et,
                latest,
            )
            # If we have something to merge, add to actions
            if with_obj:
                merged_obj = merge_resource(
                    target, target_name, with_obj, merge_network=merge_network
                )
                if merged_obj:
                    handle_output(
                        output, output_dest, merged_obj, full_diff=full_diff
                    )
            elif with_obj is False:
                continue
            else:
                merge_obj = __nothing_to_merge_with(
                    target_name, target, latest
                )
                if merge_obj:
                    handle_output(
                        output, output_dest, merge_obj, full_diff=full_diff
                    )
    elif policy_target:
        if output_to_file:
            output_dest = lib.OUTPUT_DEST_FILE
        else:
            output_dest = lib.OUTPUT_DEST_API
        policy_target = sorted(policy_target)
        if ALL in policy_target:
            for target in POLICIES:
                t_name = lib.get_metadata_name(target)
                t_uid = target[lib.METADATA_FIELD][lib.METADATA_UID_FIELD]
                target_name = f"applied policy '{t_name} - {t_uid}'"
                with_obj = get_with_obj(
                    target,
                    target_name,
                    with_file,
                    with_policy,
                    st,
                    et,
                    latest,
                    output_dest,
                )
                # If we have something to merge, add to actions
                if with_obj:
                    merged_obj = merge_resource(
                        target,
                        target_name,
                        with_obj,
                        merge_network=merge_network,
                    )
                    if merged_obj:
                        handle_output(
                            output,
                            output_dest,
                            merged_obj,
                            full_diff=full_diff,
                        )
                elif with_obj is False:
                    continue
                else:
                    merge_obj = __nothing_to_merge_with(
                        target_name, target, latest
                    )
                    if merge_obj:
                        handle_output(
                            output, output_dest, merge_obj, full_diff=full_diff
                        )
        else:
            targets = {}
            for pol_name_or_uid in policy_target:
                policies = filt.filter_obj(
                    POLICIES,
                    [
                        [lib.METADATA_FIELD, lib.NAME_FIELD],
                        [lib.METADATA_FIELD, lib.METADATA_UID_FIELD],
                    ],
                    pol_name_or_uid,
                )
                if len(policies) == 0:
                    cli.try_log(
                        "Unable to locate policy with name or UID"
                        f" {pol_name_or_uid}",
                        is_warning=True,
                    )
                    continue
                for policy in policies:
                    pol_uid = policy[lib.METADATA_FIELD][
                        lib.METADATA_UID_FIELD
                    ]
                    targets[pol_uid] = policy
            targets = sorted(
                list(targets.values()),
                key=lambda x: x[lib.METADATA_FIELD][lib.METADATA_NAME_FIELD],
            )
            if len(targets) > 0:
                pager = True
            for target in targets:
                t_name = lib.get_metadata_name(target)
                t_uid = target.get(lib.METADATA_FIELD, {}).get(
                    lib.METADATA_UID_FIELD
                )
                target_name = f"applied policy '{t_name} - {t_uid}'"
                with_obj = get_with_obj(
                    target,
                    target_name,
                    with_file,
                    with_policy,
                    st,
                    et,
                    latest,
                    output_dest,
                )
                if with_obj:
                    merged_obj = merge_resource(
                        target,
                        target_name,
                        with_obj,
                        merge_network=merge_network,
                    )
                    if merged_obj:
                        handle_output(
                            output,
                            output_dest,
                            merged_obj,
                            pager,
                            full_diff=full_diff,
                        )
                elif with_obj is False:
                    continue
                else:
                    merge_obj = __nothing_to_merge_with(
                        target_name, target, latest
                    )
                    if merge_obj:
                        handle_output(
                            output,
                            output_dest,
                            merge_obj,
                            pager,
                            full_diff=full_diff,
                        )
    else:
        cli.err_exit("No target(s) to merge.")


def get_with_obj(
    target: Dict,
    target_name: str,
    with_file: IO,
    with_policy: str,
    st,
    et,
    latest,
    dest: str = "",
) -> Optional[Union[Dict, List[Dict], bool]]:
    target_uid = target.get(lib.METADATA_FIELD, {}).get(lib.METADATA_UID_FIELD)
    if dest == lib.OUTPUT_DEST_API:
        apply_disclaimer = (
            " (Note: you will have a chance to review any merge changes before"
            " applying them.)"
        )
    else:
        apply_disclaimer = ""
    if with_file:
        with_obj = load_with_file(with_file)
        if not cli.query_yes_no(
            f"Merge {target_name} with local file '{with_file.name}'?"
            f"{apply_disclaimer}"
        ):
            return False
    elif with_policy == MATCHING:
        if not target_uid:
            cli.try_log(
                f"{target_name} has no uid, unable to match with"
                " policy stored in the Spyderbat backend.",
                is_warning=True,
            )
            return False
        with_obj = get_with_policy(target_uid, POLICIES)
        if with_obj:
            pol_name = lib.get_metadata_name(with_obj)
            pol_uid = with_obj.get(lib.METADATA_FIELD, {}).get(
                lib.METADATA_UID_FIELD
            )
            if not cli.query_yes_no(
                f"Merge {target_name} with data from applied policy "
                f"'{pol_name} - {pol_uid}'?{apply_disclaimer}"
            ):
                return False
    elif with_policy:
        with_obj = get_with_policy(with_policy, POLICIES)
        if with_obj:
            pol_name = lib.get_metadata_name(with_obj)
            pol_uid = with_obj.get(lib.METADATA_FIELD, {}).get(
                lib.METADATA_UID_FIELD
            )
            if not cli.query_yes_no(
                f"Merge {target_name} with data from applied policy"
                f" '{pol_name} - {pol_uid}'?{apply_disclaimer}"
            ):
                return False
    else:
        if latest:
            st = get_latest_timestamp(target)
        if not cli.query_yes_no(
            f"Merge {target_name} with Deviations from"
            f" {lib.epoch_to_zulu(st)} to {lib.epoch_to_zulu(et)}?"
            f"{apply_disclaimer}"
        ):
            return False
        uid = target[lib.METADATA_FIELD][lib.METADATA_UID_FIELD]
        with_obj = get_with_deviations(uid, st, et)
    return with_obj


def load_target_file(target_file: IO) -> Dict:
    rv = lib.load_resource_file(target_file)
    return rv


def get_target_policy(target_uid) -> Optional[Dict]:
    rv = p.get_policy_by_uid(target_uid, POLICIES)
    return rv


def load_with_file(with_file: IO) -> Dict:
    rv = lib.load_resource_file(with_file)
    return rv


def get_latest_timestamp(target: Dict) -> float:
    latest_timestamp = target.get(lib.METADATA_FIELD, {}).get(
        lib.LATEST_TIMESTAMP_FIELD
    )
    if latest_timestamp is not None:
        st = lib.time_inp(latest_timestamp)
    else:
        cli.err_exit(
            f"No {lib.LATEST_TIMESTAMP_FIELD} found in provided"
            f" resource {lib.METADATA_FIELD} field. Defaulting to"
            " 24hrs."
        )
    return st


def get_with_deviations(uid: str, st, et) -> List[Dict]:
    deviations = dev.get_unique_deviations(uid, st, et)
    return deviations


def filter_fingerprints(target, fingerprints) -> List[Dict]:
    filters = lib.selectors_to_filters(target)
    rv = filt.filter_fingerprints(
        fingerprints, **filters, use_context_filters=False
    )
    return rv


def get_with_policy(pol_uid: str, policies: List[Dict]) -> Dict:
    return p.get_policy_by_uid(pol_uid, policies)


def merge_resource(
    target: Dict,
    target_name: str,
    with_obj: Union[Dict, List[Dict]],
    src_cmd="merge",
    merge_network=True,
    ctx: cfgs.Context = None,
    latest=False,
    check_irrelevant=False,
) -> Optional[m_lib.MergeObject]:
    if target == with_obj:
        cli.try_log(
            f"{src_cmd} target and with-object are the same.. skipping"
        )
        return None
    if not ctx:
        ctx = cfgs.get_current_context()
    merge_with_objects = []
    if isinstance(with_obj, Dict):
        merge_with_objects = r_lib.handle_input_data(with_obj, ctx)
    else:
        for obj in with_obj:
            merge_with_objects.extend(
                r_lib.handle_input_data(data=obj, ctx=ctx)
            )
    resrc_kind = target.get(lib.KIND_FIELD)
    merge_obj = get_merge_object(resrc_kind, target, merge_network, src_cmd)
    if isinstance(merge_with_objects, list):
        for w_obj in merge_with_objects:
            if is_type_mismatch(target, target_name, src_cmd, w_obj):
                continue
            try:
                merge_obj.asymmetric_merge(w_obj, check_irrelevant)
            except m_lib.InvalidMergeError as e:
                cli.try_log(
                    f"Unable to {src_cmd} with invalid object. {w_obj}",
                    *e.args,
                )
    else:
        raise ValueError(
            f"Bug found, attempting to {src_cmd} with invalid object"
        )
    if target[lib.SPEC_FIELD] == merge_obj.get_obj_data().get(lib.SPEC_FIELD):
        if latest and src_cmd == "merge":
            cli.try_log(
                f"{src_cmd} of {target_name} produced no updates to the"
                f" '{lib.SPEC_FIELD}'.. updating {lib.LATEST_TIMESTAMP_FIELD}"
                " field to now."
            )
            merge_obj.update_latest_timestamp()
            return merge_obj
        else:
            cli.try_log(
                f"{src_cmd} of {target_name} produced no updates to the"
                f" '{lib.SPEC_FIELD}' field.. skipping"
            )
        if lib.API_CALL:
            return merge_obj
        return None
    return merge_obj


def get_merge_object(
    resrc_kind: str, target: Dict, merge_network: bool, src_cmd: str
):
    if resrc_kind == lib.BASELINE_KIND:
        merge_obj = m_lib.MergeObject(
            target,
            b.BASELINE_MERGE_SCHEMAS,
            schemas.valid_object,
            merge_network,
        )
    elif resrc_kind == lib.POL_KIND:
        resrc_type = target[lib.METADATA_FIELD][lib.METADATA_TYPE_FIELD]
        if resrc_type == lib.POL_TYPE_TRACE:
            merge_schemas = sp.T_S_POLICY_MERGE_SCHEMAS
        else:
            merge_schemas = p.POLICY_MERGE_SCHEMAS
        merge_obj = m_lib.MergeObject(
            target, merge_schemas, schemas.valid_object, merge_network
        )
    else:
        cli.try_log(
            f"The '{src_cmd}' command is not supported for {resrc_kind}",
            is_warning=True,
        )
    return merge_obj


def is_type_mismatch(
    target: Dict, target_name: str, src_cmd: str, with_obj: Dict
) -> bool:
    resrc_type = target[lib.METADATA_FIELD][lib.METADATA_TYPE_FIELD]
    with_type = with_obj[lib.METADATA_FIELD][lib.METADATA_TYPE_FIELD]
    with_kind = with_obj.get(lib.KIND_FIELD)
    target_kind = target.get(lib.KIND_FIELD)
    if target_kind == lib.POL_KIND and with_kind == lib.DEVIATION_KIND:
        pol_uid = target[lib.METADATA_FIELD].get(lib.METADATA_UID_FIELD)
        dev_pol_uid = with_obj[lib.METADATA_FIELD]["policy_uid"]
        if pol_uid != dev_pol_uid:
            cli.try_log(
                f"Error uid mismatch. Trying to {src_cmd} '{target_name}' of"
                f" type '{resrc_type}' with '{with_kind}' but the policy uids"
                " do not match. Skipping...",
                is_warning=True,
            )
            return True
        return False
    if resrc_type != with_type:
        cli.try_log(
            f"Error type mismatch. Trying to {src_cmd} '{target_name}' of type"
            f" '{resrc_type}' with '{with_kind}' object of type '{with_type}'."
            " Skipping...",
            is_warning=True,
        )
        return True
    return False


def handle_output(
    output_format: str,
    output_dest: str,
    merge_obj: m_lib.MergeObject,
    pager=False,
    full_diff=False,
):
    data = merge_obj.get_obj_data()
    if output_dest == lib.OUTPUT_DEST_API:
        apply_merge(merge_obj)
    elif output_dest == lib.OUTPUT_DEST_FILE:
        save_merge_to_file(merge_obj, output_format)
    else:
        if pager:
            cli.show(data, output_format, dest=lib.OUTPUT_DEST_PAGER)
        else:
            cli.show(data, output_format)


def apply_merge(merge_obj: m_lib.MergeObject, full_diff=False):
    data = merge_obj.get_obj_data()
    pol_name = lib.get_metadata_name(data)
    pol_uid = data[lib.METADATA_FIELD][lib.METADATA_UID_FIELD]
    if (YES_EXCEPT or not cli.YES_OPTION) and cli.query_yes_no(
        f"Review merge updates to '{pol_name}-{pol_uid}'?",
        default="no",
        ignore_yes_option=YES_EXCEPT,
    ):
        cli.show(
            merge_obj.get_diff(full_diff),
            lib.OUTPUT_RAW,
            dest=lib.OUTPUT_DEST_PAGER,
        )
    if not cli.query_yes_no(
        f"Apply merge changes to '{pol_name}-{pol_uid}'?",
        default=None,
        ignore_yes_option=YES_EXCEPT,
    ):
        return
    apply.handle_apply_policy(data)


def save_merge_to_file(merge_obj: m_lib.MergeObject, output_format):
    data = merge_obj.get_obj_data()
    pol_name = lib.get_metadata_name(data)
    pol_uid = data[lib.METADATA_FIELD].get(lib.METADATA_UID_FIELD)
    pol_str = f"{pol_name}-{pol_uid}" if pol_uid else pol_name
    if (YES_EXCEPT or not cli.YES_OPTION) and cli.query_yes_no(
        f"Review merge updates to '{pol_str}'?",
        default="no",
        ignore_yes_option=YES_EXCEPT,
    ):
        cli.show(
            merge_obj.get_diff(), lib.OUTPUT_RAW, dest=lib.OUTPUT_DEST_PAGER
        )
    if not cli.query_yes_no(
        f"Apply merge changes to '{pol_name}-{pol_uid}'?",
        default=None,
        ignore_yes_option=YES_EXCEPT,
    ):
        return
    out_fn = lib.find_resource_filename(data, "merge_output")
    out_fn = lib.unique_fn(out_fn, output_format)
    cli.show(data, output_format, dest=lib.OUTPUT_DEST_FILE, output_fn=out_fn)


def __nothing_to_merge_with(
    name: str, target, latest, src_cmd="merge"
) -> Optional[m_lib.MergeObject]:
    if latest:
        cli.try_log(
            f"{name.capitalize()} has nothing to {src_cmd} with. Updating"
            f" '{lib.LATEST_TIMESTAMP_FIELD}' field."
        )
        merge_object = get_merge_object(
            target[lib.KIND_FIELD], target, True, src_cmd
        )
        merge_object.update_latest_timestamp()
        return merge_object
    else:
        cli.try_log(
            f"{name.capitalize()} has nothing to {src_cmd} with... skipping."
        )



import json
import yaml
from typing import Dict, List, Optional, Tuple

import spyctl.config.configs as cfg
import spyctl.resources.api_filters as af
import spyctl.schemas_v2 as schemas
import spyctl.spyctl_lib as lib
from spyctl import api

NS_LABEL = "kubernetes.io/metadata.name"


class RulesObject:
    def __init__(self, verb: str) -> None:
        self.verb = verb


class NamespaceLabels:
    def __init__(self, namespace_labels: List[str] = []):
        self.namespace_labels = namespace_labels

    def as_dict(self):
        rv = {
            lib.NAMESPACE_SELECTOR_FIELD: {
                lib.MATCH_LABELS_FIELD: {NS_LABEL: self.namespace_labels}
            }
        }
        return rv

    def add_namespace(self, namespace: str):
        self.namespace_labels.append(namespace)

    @property
    def first_namespace(self) -> Optional[str]:
        if len(self.namespace_labels) == 0:
            return None
        return self.namespace_labels[0]

    def __eq__(self, __value: object) -> bool:
        if not isinstance(__value, __class__):
            return False
        return set(self.namespace_labels) == set(__value.namespace_labels)


class DetectedImage:
    def __init__(self, image: str):
        self.image = image
        self.namespaces = set()

    def add_namespace(self, namespace: Optional[str]):
        self.namespaces.add(namespace)

    def namespaces_match(self, other_namespaces: set) -> bool:
        return self.namespaces == other_namespaces


class ContainerRules(RulesObject):
    def __init__(self, verb="allow", include_namespaces=False):
        super().__init__(verb)
        self.images: Dict[str, DetectedImage] = {}
        self.include_namespaces = include_namespaces

    def add_container(self, container: Dict):
        image = container["image"]
        namespaces_labels: Dict = container.get("pod_namespace_labels", {})
        namespace = namespaces_labels.get(NS_LABEL, container.get("namespace"))
        if not namespace:
            lib.try_log(
                f"Container {container['container_name']} with from {image} has no namespace.. skipping",  # noqa: E501
                is_warning=True,
            )
            return
        self.__add_image(image, namespace)

    def as_list(self):
        rv = self.__aggregate_images()
        return rv

    def __aggregate_images(self):
        if self.include_namespaces:
            rv = self.__agg_images_by_ns()
        else:
            rv = [
                {
                    "verb": self.verb,
                    lib.IMAGE_FIELD: sorted(
                        sorted([image.image for image in self.images.values()])
                    ),
                }
            ]
        return rv

    def __agg_images_by_ns(self):
        rv = []
        agg: Dict[Tuple, List[str]] = {}
        for image in self.images.values():
            namespaces = tuple(sorted(image.namespaces))
            agg.setdefault(namespaces, [])
            agg[namespaces].append(image.image)
        for namespaces, images in agg.items():
            if len(namespaces) == 1:
                ns_selector = schemas.NamespaceSelectorModel(
                    matchLabels={"kubernetes.io/metadata.name": namespaces[0]}
                )
            else:
                expr = schemas.SelectorExpression(
                    key="kubernetes.io/metadata.name",
                    operator=lib.IN_OPERATOR,
                    values=namespaces,
                )
                ns_selector = schemas.NamespaceSelectorModel(
                    matchExpressions=[
                        expr.dict(by_alias=True, exclude_unset=True)
                    ]
                )
            rv.append(
                {
                    lib.NAMESPACE_SELECTOR_FIELD: ns_selector.dict(
                        by_alias=True, exclude_unset=True
                    ),
                    "verb": self.verb,
                    lib.IMAGE_FIELD: sorted(images),
                }
            )

        def sort_key(item: Dict):
            labels = item[lib.NAMESPACE_SELECTOR_FIELD].get(
                lib.MATCH_LABELS_FIELD
            )
            if labels:
                namespace = labels.get(NS_LABEL)
                return namespace
            return ""
            # expressions = item[lib.NAMESPACE_SELECTOR_FIELD].get(
            # if isinstance(namespaces, str) else namespaces[0]

        rv.sort(key=sort_key)
        return rv

    def __add_image(self, image, namespace):
        dti = self.images.setdefault(image, DetectedImage(image))
        dti.add_namespace(namespace)


class ClusterRuleset:
    def __init__(self, name: str, cluster: str = None):
        self.name = name
        self.rules: Dict[str, Dict] = {}  # verb -> type -> RulesObject
        self.cluster = cluster

        if not name:
            if cluster:
                self.name = f"{cluster}-cluster-ruleset"
            else:
                self.name = "default-ruleset"

    def add_rules(
        self, verb: str, rules_type: str, include_namespaces: bool
    ) -> RulesObject:
        rules = self.rules.get(verb, {}).get(rules_type)
        if rules:
            return rules
        if rules_type == lib.RULES_TYPE_CONTAINER:
            rules = ContainerRules(verb, include_namespaces)
            self.rules.setdefault(verb, {})[rules_type] = rules
            return rules
        else:
            lib.err_exit(f"Unknown rules type: {rules_type}")

    def as_dict(self):
        return self.__as_cluster_ruleset_dict()

    def __as_cluster_ruleset_dict(self):
        if not self.cluster:
            lib.err_exit("Cluster name or UID is required for cluster ruleset")
        rv = {
            lib.API_FIELD: lib.API_VERSION,
            lib.KIND_FIELD: lib.CLUSTER_RULESET_RESOURCE.kind,
            lib.METADATA_FIELD: {
                lib.NAME_FIELD: self.name,
                lib.TYPE_FIELD: lib.RULESET_TYPE_CLUS,
            },
            lib.SPEC_FIELD: {
                lib.RULES_FIELD: self.__compile_rules(),
            },
        }
        return rv

    def __compile_rules(self) -> Dict:
        rv = {}
        for _, rules in self.rules.items():
            for rules_type, rules_obj in rules.items():
                rv.setdefault(rules_type, []).extend(rules_obj.as_list())
        rv = {k: rv[k] for k in sorted(rv) if rv[k] != []}
        return rv


def create_blank_ruleset(name: str):
    pass


def create_ruleset(
    name: str, generate_rules: bool, time, **filters
) -> ClusterRuleset:
    ruleset = ClusterRuleset(name, filters.get(lib.CLUSTER_OPTION))
    if generate_rules:
        generate_cluster_ruleset(
            ruleset, lib.CLUSTER_RULESET_RULE_TYPES, time, **filters
        )
    return ruleset


def generate_cluster_ruleset(
    ruleset: ClusterRuleset, rule_types, time, **filters
):
    cluster = filters.get(lib.CLUSTER_OPTION)
    if not cluster:
        lib.err_exit("Cluster name or UID is required for cluster ruleset")
    ruleset.cluster = cluster
    if not rule_types:
        rule_types = lib.CLUSTER_RULESET_RULE_TYPES
    for rule_type in rule_types:
        if rule_type == lib.RULES_TYPE_CONTAINER:
            generate_container_rules(ruleset, time, **filters)
    return ruleset


def generate_container_rules(ruleset: ClusterRuleset, time, **filters):
    filters = filters.copy()
    include_namespaces = False
    if namespaces := filters.get(lib.NAMESPACE_OPTION, []):
        include_namespaces = True
        if "__all__" in namespaces:
            # We don't need to filter for specific namespaces at this point
            filters.pop(lib.NAMESPACE_OPTION)
    ctx = cfg.get_current_context()
    sources, filters = af.Containers.build_sources_and_filters(**filters)
    pipeline = af.Containers.generate_pipeline(filters=filters)
    lib.try_log("Generating container rules...")
    container_rules: ContainerRules = ruleset.add_rules(
        "allow", lib.RULES_TYPE_CONTAINER, include_namespaces
    )
    for container in api.get_containers(
        *ctx.get_api_data(), sources, time, pipeline, limit_mem=True
    ):
        container_rules.add_container(container)



import os
import shutil
import unittest.mock as mock
from fnmatch import fnmatch
from itertools import groupby
from pathlib import Path
from typing import Tuple

import pytest
from click.testing import CliRunner

import spyctl.tests.mock_functions as mock_func
from spyctl import spyctl
from spyctl.config.configs import CURR_CONTEXT_NONE, set_testing
from spyctl.spyctl_lib import time_inp

API_KEY = os.environ.get("API_KEY")
API_URL = os.environ.get("API_URL")
ORG = os.environ.get("ORG")
TEST_SECRET = "__test_secret__"
TEST_CONTEXT = "__test_context__"
CURR_PATH = Path.cwd()
WORKSPACE_DIR_NAME = "test_workspace__"
WORKSPACE_PATH = Path(f"./{WORKSPACE_DIR_NAME}")
WORKSPACE_CONFIG = Path(str(WORKSPACE_PATH.absolute()) + "/.spyctl/config")
CURRENT_CONTEXT = None


class SetupException(Exception):
    pass


def test_version():
    runner = CliRunner()
    response = runner.invoke(spyctl.main, ["-v"])
    test_pat = "Spyctl, version *.*.*"
    assert response.exit_code == 0
    assert fnmatch(response.output, test_pat)
    response = runner.invoke(spyctl.main, ["--version"])
    test_pat = "Spyctl, version *.*.*"
    assert response.exit_code == 0
    assert fnmatch(response.output, test_pat)


def test_help():
    runner = CliRunner()
    response = runner.invoke(spyctl.main, ["-h"])
    assert response.exit_code == 0
    response = runner.invoke(spyctl.main, ["--help"])
    assert response.exit_code == 0
    response = runner.invoke(spyctl.main, ["get", "redflags", "--help"])
    assert response.exit_code == 0
    assert "Options for redflags:" in response.output
    response = runner.invoke(spyctl.main, ["config", "--help"])
    assert response.exit_code == 0


@mock.patch(
    "spyctl.commands.get.api.get_namespaces", mock_func.mock_get_namespaces
)
def test_get_namespaces():
    get_resource("namespaces", TWOHOURS)
    get_resource("namespaces", TWOHOURS + OUTYML)


TWOHOURS = ["-t", "2h"]
OUTYML = ["-o", "yaml"]
MINS_30 = ["-t", "30m"]
MINS_5 = ["-t", "5m"]


def remove_timestamps(str_list):
    rv = []
    for string in str_list:
        try:
            time_inp(string)
        except ValueError:
            rv.append(string)
    return rv


# asserts that the table has some population and returns
# name_or_id values that should give the first line again
def process_table_response(output: str):
    name_or_id_fields = ("NAME", "UID", "FLAG", "IMAGE", "CGROUP")
    lines = output.strip().splitlines()
    assert len(lines) > 1
    header_line, first_line = lines[:2]
    headers = []
    for non_empty_group, chars in groupby(
        enumerate(header_line), lambda x: not x[1].isspace()
    ):
        if non_empty_group:
            pos, header = next(chars)
            header += "".join(c for _, c in chars)
            headers.append((pos, header))
    rv = []
    for i, (pos, header) in enumerate(headers):
        if header in name_or_id_fields:
            next_pos = headers[i + 1][0] if i < len(headers) - 1 else None
            table_element = first_line[pos:next_pos].strip()
            rv.append(table_element)
    print(header_line)
    print(headers)
    assert len(rv) > 0
    comparison_line = remove_timestamps(first_line.split())
    return comparison_line, rv


resources = (
    "clusters",
    "sources",
    "policies",
    "pods",
    "nodes",
    "deployments",
    "fingerprints",
    "namespaces",  # doesn't output a table
    "redflags",
    "opsflags",
)


@mock.patch.multiple(
    "spyctl.commands.get.api",
    get_clusters=mock_func.mock_get_clusters,
    get_deployments=mock_func.mock_get_deployments,
    get_guardian_fingerprints=mock_func.mock_get_guardian_fingerprints,
    get_namespaces=mock_func.mock_get_namespaces,
    get_nodes=mock_func.mock_get_nodes,
    get_opsflags=mock_func.mock_get_opsflags,
    get_pods=mock_func.mock_get_pods,
    get_policies=mock_func.mock_get_policies,
    get_redflags=mock_func.mock_get_redflags,
    get_sources=mock_func.mock_get_sources,
)
@mock.patch.multiple(
    "spyctl.resources.deployments.api",
    get_deployments=mock_func.mock_get_deployments,
)
@mock.patch.multiple(
    "spyctl.resources.fingerprints.api",
    get_guardian_fingerprints=mock_func.mock_get_guardian_fingerprints,
)
@mock.patch.multiple(
    "spyctl.resources.flags.api",
    get_opsflags=mock_func.mock_get_opsflags,
    get_redflags=mock_func.mock_get_redflags,
)
@mock.patch.multiple(
    "spyctl.resources.namespaces.api",
    get_namespaces=mock_func.mock_get_namespaces,
)
@mock.patch.multiple(
    "spyctl.resources.nodes.api",
    get_nodes=mock_func.mock_get_nodes,
)
@mock.patch.multiple(
    "spyctl.resources.pods.api",
    get_pods=mock_func.mock_get_pods,
)
@pytest.mark.parametrize("resource", resources)
def test_get_resources(resource):
    args = get_args(resource)
    if resource not in ["opsflags", "fingerprints"]:
        time_range = MINS_30
    else:
        time_range = TWOHOURS
    output = OUTYML
    table = get_resource(resource, args + time_range)
    comparison_line, ids = process_table_response(table)
    for name_or_id in ids:
        single = get_resource(resource, args + time_range + [name_or_id])
        first_output = remove_timestamps(single.splitlines()[1].split())
        assert first_output == comparison_line
    time_range = MINS_5
    get_resource(resource, args + time_range + output)


def get_resource(resource, args=[], print_output=False):
    runner = CliRunner(mix_stderr=False)
    response = runner.invoke(spyctl.main, ["get", resource, *args])
    if print_output:
        print(response.output)
    assert response.exit_code == 0
    return response.stdout


SPECIAL_GET_ARGS = {"fingerprints": ["--type", "linux-service"]}


def get_args(resource: str) -> Tuple:
    return SPECIAL_GET_ARGS.get(resource, [])


resources_dir = Path(__file__).parent / "test_resources"


def test_create():
    runner = CliRunner()
    response = runner.invoke(
        spyctl.main,
        ["create", "policy", "-f", resources_dir / "test_baseline.yaml"],
    )
    assert response.exit_code == 0
    with open(resources_dir / "test_policy.yaml", "r", encoding="utf-8") as f:
        assert response.output.strip("\n") == f.read().strip("\n")
    response = runner.invoke(
        spyctl.main,
        ["create", "policy", "-f", resources_dir / "test_fprint_group.yaml"],
    )
    assert response.exit_code == 0
    with open(
        resources_dir / "test_created_policy.yaml", "r", encoding="utf-8"
    ) as f:
        assert response.output.strip("\n") == f.read().strip("\n")


@mock.patch(
    "spyctl.commands.get.api.get_policies", mock_func.mock_get_policies
)
@mock.patch.multiple(
    "spyctl.commands.apply.api",
    get_policies=mock_func.mock_get_policies,
    delete_policy=mock_func.mock_delete_policy,
    post_new_policy=mock_func.mock_post_new_policy,
    put_policy_update=mock_func.mock_put_policy_update,
)
@mock.patch.multiple(
    "spyctl.commands.delete.api",
    get_policies=mock_func.mock_get_policies,
    delete_policy=mock_func.mock_delete_policy,
)
def test_update_policy():
    runner = CliRunner()
    response = runner.invoke(
        spyctl.main, ["apply", "-f", resources_dir / "test_policy.yaml"]
    )
    assert response.exit_code == 0
    assert response.output.startswith(
        "Successfully applied new container guardian policy with uid:"
    )
    response = runner.invoke(
        spyctl.main, ["get", "policies", "spyderbat-test"]
    )
    assert response.exit_code == 0
    assert "spyderbat-test" in response.output
    lines = response.stdout.splitlines()
    uid = lines[2].split(" ")[0]
    get_runner = CliRunner(mix_stderr=False)
    response = get_runner.invoke(
        spyctl.main, ["get", "policies", uid, "-o", "yaml"]
    )
    assert response.exit_code == 0
    assert "spyderbat-test" in response.output
    policy_with_uid = "uid_pol.yaml"
    with open(policy_with_uid, "w") as f:
        f.write(response.output)
    response = runner.invoke(spyctl.main, ["apply", "-f", policy_with_uid])
    assert response.exit_code == 0
    assert response.output.startswith("Successfully updated policy")
    response = runner.invoke(
        spyctl.main, ["delete", "policy", "-y", "spyderbat-test"]
    )
    assert response.exit_code == 0
    assert response.output.startswith("Successfully deleted policy")


@mock.patch(
    "spyctl.commands.get.api.get_policies", mock_func.mock_get_policies
)
@mock.patch.multiple(
    "spyctl.commands.apply.api",
    get_policies=mock_func.mock_get_policies,
    delete_policy=mock_func.mock_delete_policy,
    post_new_policy=mock_func.mock_post_new_policy,
)
@mock.patch.multiple(
    "spyctl.commands.delete.api",
    get_policies=mock_func.mock_get_policies,
    delete_policy=mock_func.mock_delete_policy,
)
def test_apply_delete():
    runner = CliRunner()
    response = runner.invoke(
        spyctl.main, ["apply", "-f", Path(resources_dir, "test_policy.yaml")]
    )
    assert response.exit_code == 0
    assert response.output.startswith(
        "Successfully applied new container guardian policy with uid:"
    )
    response = runner.invoke(
        spyctl.main, ["get", "policies", "spyderbat-test"]
    )
    assert response.exit_code == 0
    assert "spyderbat-test" in response.output
    response = runner.invoke(
        spyctl.main, ["delete", "policy", "-y", "spyderbat-test"]
    )
    assert response.exit_code == 0
    assert response.output.startswith("Successfully deleted policy")


def test_diff():
    runner = CliRunner()
    response = runner.invoke(
        spyctl.main,
        [
            "diff",
            "-f",
            str(resources_dir) + "/test_baseline.yaml",
            "-w",
            str(resources_dir) + "/test_baseline_extra.yaml",
            "-y",
        ],
    )
    print(response.output)
    assert response.exit_code == 0
    assert "+     - name: python3.7" in response.output


def test_merge():
    runner = CliRunner()
    response = runner.invoke(
        spyctl.main,
        [
            "merge",
            "-f",
            str(resources_dir) + "/test_baseline.yaml",
            "-w",
            str(resources_dir) + "/test_baseline_extra.yaml",
            "-y",
        ],
    )
    print(response.output)
    assert response.exit_code == 0
    assert "\n    - name: python3.7" in response.output
    assert "\n    - name: sh" in response.output


def env_setup() -> bool:
    # Tests that the proper environment variables exist in pyproject.toml
    if not API_KEY or API_KEY == "__NONE__":
        print(
            "No api key provided. Edit the API_KEY environment variable in"
            " pyproject.toml"
        )
        return False
    if not API_URL or API_URL == "__NONE__":
        print(
            "No api url provided. Edit the API_URL environment variable in"
            ' pyproject.toml to "API_URL=https://api.spyderbat.com"'
        )
        return False
    if not ORG or ORG == "__NONE__":
        print(
            "No organization provided. Edit the ORG environment variable in"
            " pyproject.toml"
        )
        return False
    return True


def create_secret():
    runner = CliRunner()
    result = runner.invoke(
        spyctl.main,
        [
            "config",
            "set-apisecret",
            "-k",
            API_KEY,
            "-u",
            API_URL,
            TEST_SECRET,
        ],
    )
    if result.exit_code != 0:
        raise SetupException("Unable to create test secret")


def delete_secret():
    runner = CliRunner()
    runner.invoke(
        spyctl.main,
        [
            "config",
            "delete-apisecret",
            "-y",
            TEST_SECRET,
        ],
    )


def create_context(name):
    runner = CliRunner()
    result = runner.invoke(
        spyctl.main,
        [
            "config",
            "set-context",
            "-s",
            TEST_SECRET,
            "-o",
            ORG,
            "-g",
            name,
        ],
    )
    if result.exit_code != 0:
        raise SetupException("Unable to create test context")


def delete_context(name):
    runner = CliRunner()
    runner.invoke(
        spyctl.main,
        ["config", "delete-context", name],
    )


def delete_test_policies():
    runner = CliRunner()
    runner.invoke(spyctl.main, ["delete", "policy", "-y", "spyderbat-test"])


def current_context():
    runner = CliRunner()
    result = runner.invoke(
        spyctl.main,
        [
            "config",
            "current-context",
        ],
    )
    if result.exit_code != 0:
        raise SetupException("Unable get current context")
    current_ctx = result.output.strip("\n")
    return current_ctx


def use_test_context(name):
    runner = CliRunner()
    runner.invoke(spyctl.main, ["config", "use-context", name])


def use_current_context():
    if CURRENT_CONTEXT:
        runner = CliRunner()
        runner.invoke(spyctl.main, ["config", "use-context", CURRENT_CONTEXT])


def setup_module():
    if not env_setup():
        raise SetupException("Check environment variables in pyproject.toml")
    global CURRENT_CONTEXT
    current_ctx = current_context()
    if current_ctx and current_ctx != CURR_CONTEXT_NONE:
        CURRENT_CONTEXT = current_ctx
    if WORKSPACE_PATH.exists():
        teardown_module()
        raise SetupException("Workspace path already exists")
    WORKSPACE_PATH.mkdir(exist_ok=False)
    os.chdir(WORKSPACE_PATH)
    create_secret()
    create_context(TEST_CONTEXT)
    use_test_context(TEST_CONTEXT)
    # Set a global variable in configs.py to force reloading
    # the config file. Otherwise the global variable LOADED_CONFIG
    # remains set across invocations
    set_testing()


def teardown_module():
    delete_test_policies()
    delete_context(TEST_CONTEXT)
    delete_secret()
    use_current_context()
    os.chdir(CURR_PATH)
    shutil.rmtree(WORKSPACE_PATH, ignore_errors=True)



"""
Contains the ClusterPolicy class, which is used to define a policy for a
kubernetes cluster.
"""

from typing import Dict, List

import spyctl.resources.cluster_rulesets as crs
import spyctl.resources.clusters as c
import spyctl.schemas_v2 as schemas
import spyctl.spyctl_lib as lib


def create_cluster_policy(
    name: str,
    mode: str,
    st: float,
    et: float,
    no_rs_gen: bool,
    clusters: List[str] = None,
    namespace: List[str] = None,
) -> Dict:
    """
    Create a cluster policy and output it to stdout in the desired format.

    Args:
        name (str): The name of the cluster policy.
        mode (str): The mode of the cluster policy.
        st (str): The start time of the cluster policy.
        et (str): The end time of the cluster policy.
        no_rs_gen (bool): Whether or not to generate rulesets.
        cluster (str, optional): The cluster name. Defaults to None.
        namespace (List[str], optional): Scope ruleset rules to namespace(s).
    """
    metadata = schemas.GuardianMetadataModel(name=name, type=lib.POL_TYPE_CLUS)
    rulesets: List[crs.ClusterRuleset] = []
    spec = schemas.ClusterPolicySpecModel(
        clusterSelector=schemas.ClusterSelectorModel(
            cluster="PROVIDE_CLUSTER_NAME"
        ),
        mode=mode,
        enabled=True,
        rulesets=[],
    )
    if clusters:
        lib.try_log("Validating cluster(s) exist within the system.")
        spec.cluster_selector.cluster = (
            clusters[0] if len(clusters) == 1 else clusters
        )
        for clus in clusters:
            if not c.cluster_exists(clus):
                lib.err_exit(f"Cluster {clus} does not exist")
            if no_rs_gen:
                continue
            filters = {
                "cluster": clus,
            }
            if namespace:
                filters["namespace"] = namespace
            rulesets.append(
                crs.create_ruleset(
                    f"{clus}_ruleset", True, (st, et), **filters
                )
            )
    policy = schemas.ClusterPolicyModel(
        apiVersion=lib.API_VERSION,
        kind=lib.POL_KIND,
        metadata=metadata,
        spec=spec,
    )
    if not rulesets:
        rv = policy.dict(by_alias=True, exclude_unset=True)
    else:
        items = []
        for ruleset in rulesets:
            rs_name = ruleset.name
            spec.rulesets.append(rs_name)
            items.append(ruleset.as_dict())
        pol_dict = policy.dict(by_alias=True, exclude_unset=True)
        items.append(pol_dict)
        rv = {
            lib.API_FIELD: lib.API_VERSION,
            lib.ITEMS_FIELD: items,
        }
    return rv



"""Handles the 'edit' subcommand for spyctl."""

# pylint: disable=broad-exception-caught

import json
import tempfile
from io import TextIOWrapper
from typing import IO, Callable, Dict

import click
import yaml

import spyctl.config.configs as cfg
import spyctl.resources.notification_configs as nc
import spyctl.resources.notification_targets as nt
import spyctl.resources.policies as p
import spyctl.schemas_v2 as schemas
import spyctl.spyctl_lib as lib
from spyctl import api, cli

# ----------------------------------------------------------------- #
#                          Edit Subcommand                          #
# ----------------------------------------------------------------- #


@click.command("edit", cls=lib.CustomCommand, epilog=lib.SUB_EPILOG)
@click.help_option("-h", "--help", hidden=True)
@click.argument("resource", type=lib.EditResourcesParam(), required=False)
@click.argument("name_or_id", required=False)
@click.option(
    "-f",
    "--filename",
    help="Filename to use to edit the resource.",
    metavar="",
    type=click.File(mode="r+"),
)
@click.option(
    "-y",
    "--yes",
    "--assume-yes",
    is_flag=True,
    help='Automatic yes to prompts; assume "yes" as answer to all prompts and'
    " run non-interactively.",
)
def edit(resource, name_or_id, filename, yes=False):
    """Edit resources by resource and name, or by resource and ids"""
    if yes:
        cli.set_yes_option()
    handle_edit(resource, name_or_id, filename)


# ----------------------------------------------------------------- #
#                          Edit Handlers                            #
# ----------------------------------------------------------------- #

EDIT_PROMPT = (
    "# Please edit the object below. Lines beginning with a '#' will be ignored,\n"  # noqa
    "# and an empty file will abort the edit. If an error occurs while saving this file will be\n"  # noqa
    "# reopened with the relevant failures.\n"
    "#\n"
)

KIND_TO_RESOURCE_TYPE: Dict[str, str] = {
    lib.BASELINE_KIND: lib.BASELINES_RESOURCE.name,
    lib.CONFIG_KIND: lib.CONFIG_ALIAS.name,
    lib.FPRINT_GROUP_KIND: lib.FINGERPRINT_GROUP_RESOURCE.name,
    lib.FPRINT_KIND: lib.FINGERPRINTS_RESOURCE.name,
    lib.POL_KIND: lib.POLICIES_RESOURCE.name,
    (lib.POL_KIND, lib.POL_TYPE_TRACE): lib.SUPPRESSION_POLICY_RESOURCE.name,
    lib.SECRET_KIND: lib.SECRETS_ALIAS.name,
    lib.UID_LIST_KIND: lib.UID_LIST_RESOURCE.name,
    lib.DEVIATION_KIND: lib.DEVIATIONS_RESOURCE.name,
    lib.NOTIFICATION_KIND: lib.NOTIFICATION_CONFIGS_RESOURCE.name,
    lib.TARGET_KIND: lib.NOTIFICATION_TARGETS_RESOURCE.name,
    lib.RULESET_KIND: "ruleset",
}


def handle_edit(resource=None, name_or_id=None, file: IO = None):
    """
    Handles the 'edit' command for a specified resource and name or id.

    Args:
        resource (str): The resource to edit.
        name_or_id (str): The name or id of the resource to edit.
        file (IO): The file to edit.

    Returns:
        None
    """
    if file:
        handle_edit_file(file)
    else:
        if not resource or not name_or_id:
            cli.err_exit("Must specify resource and name or id.")
        if resource == lib.CLUSTER_RULESET_RESOURCE:
            handle_edit_ruleset(name_or_id)
        elif resource == lib.NOTIFICATION_CONFIGS_RESOURCE:
            handle_edit_notif_config(name_or_id)
        elif resource == lib.NOTIFICATION_TARGETS_RESOURCE:
            handle_edit_notif_tgt(name_or_id)
        elif resource == lib.RULESETS_RESOURCE:
            handle_edit_ruleset(name_or_id)
        elif resource == lib.CONTAINER_POL_RESOURCE:
            handle_edit_policy(name_or_id, lib.POL_TYPE_CONT)
        elif resource == lib.LINUX_SVC_POL_RESOURCE:
            handle_edit_policy(name_or_id, lib.POL_TYPE_SVC)
        elif resource == lib.POLICIES_RESOURCE:
            handle_edit_policy(name_or_id)
        elif resource == lib.SUPPRESSION_POLICY_RESOURCE:
            handle_edit_policy(name_or_id, lib.POL_TYPE_TRACE)
        else:
            cli.err_exit(
                f"The 'edit' command is not supported for '{resource}'"
            )


def handle_edit_file(file: IO):
    """
    Handle editing a file.

    Args:
        file (IO): The file to be edited.

    Raises:
        ValueError: If the file is not a valid resource file.
        ValueError: If editing a resource of the given kind is not supported.
    """
    resource = lib.load_resource_file(file)
    if not isinstance(resource, Dict):
        cli.err_exit("Invalid file for editing.")
    kind = resource.get(lib.KIND_FIELD)
    if kind not in KIND_TO_RESOURCE_TYPE:
        cli.err_exit(f"Editing resource of kind '{kind}' not supported.")
    if kind == lib.POL_KIND:
        m_type = resource[lib.METADATA_FIELD][lib.METADATA_TYPE_FIELD]
        if m_type == lib.POL_TYPE_TRACE:
            kind = (lib.POL_KIND, lib.POL_TYPE_TRACE)
    m_type = resource[lib.METADATA_FIELD].get(lib.METADATA_TYPE_FIELD)
    resource_type = KIND_TO_RESOURCE_TYPE.get((kind, kind))
    if not resource_type:
        resource_type = KIND_TO_RESOURCE_TYPE[kind]
    edit_resource(
        cli.make_yaml(resource),
        file,
        resource_type,
        schemas.KIND_TO_SCHEMA[kind],
        apply_file_edits,
    )


def handle_edit_ruleset(name_or_id, rs_type=None):
    ctx = cfg.get_current_context()
    params = {
        "type": rs_type,
        "name_or_uid_contains": name_or_id,
    }
    rulesets = api.get_rulesets(*ctx.get_api_data(), params=params)
    if not rulesets:
        desc = f" {rs_type} " if rs_type else " "
        cli.err_exit(f"No{desc}rulesets matching '{name_or_id}'")
    if len(rulesets) > 1:
        cli.err_exit(f"Ruleset '{name_or_id}' is ambiguous, use full UID.")
    ruleset = rulesets[0]
    resource_yaml = cli.make_yaml(ruleset)
    edit_resource(
        resource_yaml,
        ruleset[lib.METADATA_FIELD][lib.METADATA_UID_FIELD],
        "ruleset",
        schemas.KIND_TO_SCHEMA[lib.RULESET_KIND],
        apply_ruleset_edits,
    )


def handle_edit_notif_config(name_or_id):
    """
    Edit a notification configuration based on the provided name or ID.

    Args:
        name_or_id (str): The name or ID of the notification configuration to
            edit.

    Raises:
        ValueError: If the provided name or ID is ambiguous or no matching
            notification configurations are found.
        ValueError: If editing legacy notification configurations is not
            supported.

    Returns:
        None
    """
    ctx = cfg.get_current_context()
    notif_pol = api.get_notification_policy(*ctx.get_api_data())
    routes = notif_pol.get(lib.ROUTES_FIELD)
    edit_index = None
    edit_id = None
    for i, route in enumerate(routes):
        notif_id = route.get(lib.DATA_FIELD, {}).get(lib.ID_FIELD)
        name = route.get(lib.DATA_FIELD, {}).get(lib.NAME_FIELD)
        if name_or_id in [name, notif_id]:
            if edit_index is not None and name == name_or_id:
                cli.err_exit(f"{name_or_id} is ambiguous, use ID")
            edit_index = i
            edit_id = notif_id
    if edit_index is None:
        cli.err_exit(f"No notification configs matching '{name_or_id}'.")
    config = (
        routes[edit_index]
        .get(lib.DATA_FIELD, {})
        .get(lib.NOTIF_SETTINGS_FIELD)
    )
    key = (
        lib.NOTIFICATION_KIND,
        config.get(lib.METADATA_FIELD).get(lib.METADATA_TYPE_FIELD),
    )
    if key not in schemas.KIND_TO_SCHEMA:
        key = lib.NOTIFICATION_KIND
    if not config:
        cli.err_exit("Editing legacy notification configs not supported.")
    resource_yaml = cli.make_yaml(config)
    edit_resource(
        resource_yaml,
        edit_id,
        lib.NOTIFICATION_CONFIGS_RESOURCE.name,
        schemas.KIND_TO_SCHEMA[key],
        apply_config_edits,
    )


def handle_edit_notif_tgt(name_or_id):
    """
    Handle the editing of a notification target based on the given name or ID.

    Args:
        name_or_id (str): The name or ID of the notification target to be
            edited.

    Returns:
        None
    """
    ctx = cfg.get_current_context()
    notif_pol = api.get_notification_policy(*ctx.get_api_data())
    targets: Dict = notif_pol[lib.TARGETS_FIELD]
    edit_tgt = None
    # check if name exists
    if name_or_id in targets:
        tgt_data = targets[name_or_id]
        edit_tgt = nt.Target(backend_target={name_or_id: tgt_data})
    if not edit_tgt:
        for tgt_name, tgt in targets.items():
            tgt_id = tgt.get(lib.DATA_FIELD, {}).get(lib.ID_FIELD)
            if tgt_id is None:
                continue
            if tgt_id == name_or_id:
                edit_tgt = nt.Target(backend_target={tgt_name: tgt})
                break
    if not edit_tgt:
        cli.err_exit(f"No notification targets matching '{name_or_id}'.")
    resource_yaml = cli.make_yaml(edit_tgt.as_dict())
    edit_resource(
        resource_yaml,
        edit_tgt.id,
        lib.NOTIFICATION_TARGETS_RESOURCE.name,
        schemas.NotificationTgtResourceModel,
        apply_tgt_edits,
    )


def handle_edit_policy(name_or_id, pol_type=None):
    """
    Handle the editing of a policy based on the given name or ID.

    Args:
        name_or_id (str): The name or ID of the policy to be edited.

    Returns:
        None
    """
    ctx = cfg.get_current_context()
    params = {
        "type": pol_type,
        "name_or_uid_contains": name_or_id,
    }
    policies = api.get_policies(*ctx.get_api_data(), params=params)
    if len(policies) > 1:
        cli.err_exit(f"Policy '{name_or_id}' is ambiguous, use full UID.")
    if not policies:
        desc = f" {pol_type} " if pol_type else " "
        cli.err_exit(f"No{desc}policies matching '{name_or_id}'.")
    policy = policies[0]
    resource_yaml = cli.make_yaml(policy)
    edit_resource(
        resource_yaml,
        policy[lib.METADATA_FIELD][lib.METADATA_UID_FIELD],
        lib.POLICIES_RESOURCE.name,
        schemas.KIND_TO_SCHEMA[lib.POL_KIND],
        apply_policy_edits,
    )


def edit_resource(
    resource_yaml: str,
    resource_id: str,
    resource_type,
    validator: Callable,
    apply_func: Callable,
):
    """
    Edit a resource using a YAML file.

    Args:
        resource_yaml (str): The YAML content of the resource.
        resource_id (str): The ID of the resource.
        resource_type: The type of the resource.
        validator (Callable): A function that validates the edited YAML
            content.
        apply_func (Callable): A function that applies the edited YAML content.

    Returns:
        None
    """
    temp_file = None
    while True:
        if not temp_file:
            edit_yaml = click.edit(
                __add_edit_prompt(resource_yaml), extension=".yaml"
            )
        else:
            try:
                with open(temp_file, "r", encoding="UTF-8") as f:
                    tmp_yaml = f.read()
                edit_yaml = click.edit(tmp_yaml, extension=".yaml")
            except Exception as e:
                cli.err_exit(str(e))
        if not edit_yaml or __strip_comments(edit_yaml) == resource_yaml:
            cli.try_log("Edit cancelled, no changes made.")
            exit(0)
        try:
            edit_dict = yaml.load(edit_yaml, lib.UniqueKeyLoader)
            error = None
            validator(**edit_dict)
        except Exception as e:
            error = str(e)
        if error and temp_file:
            edit_yaml = __add_error_comments(
                resource_type,
                edit_yaml,
                error,
                resource_id,
            )
            with open(temp_file, "w", encoding="UTF-8") as f:
                f.write(edit_yaml)
            cli.try_log(f"Edit failed, edits saved to {temp_file}")
            exit(1)
        if error:
            edit_yaml = __add_error_comments(
                resource_type,
                edit_yaml,
                error,
                resource_id,
            )
            temp_file = tempfile.NamedTemporaryFile(
                "w", delete=False, prefix="spyctl-edit-", suffix=".yaml"
            )
            temp_file.write(edit_yaml)
            cli.try_log(f"Edit failed, edits saved to {temp_file.name}")
            temp_file.close()
            temp_file = temp_file.name
            continue
        else:
            apply_func(edit_dict, resource_id)
            break


def apply_config_edits(edit_dict: Dict, config_id: str):
    """
    Apply edits to a notification configuration.

    Args:
        edit_dict (Dict): A dictionary containing the edits to be applied.
        config_id (str): The ID of the configuration to be edited.
    """
    ctx = cfg.get_current_context()
    notif_pol = api.get_notification_policy(*ctx.get_api_data())
    config = nc.NotificationConfig(edit_dict)
    config.id = config_id
    routes = notif_pol.get(lib.ROUTES_FIELD, [])
    edit_index = None
    for i, route in enumerate(routes):
        cfg_id = route.get(lib.DATA_FIELD, {}).get(lib.ID_FIELD)
        if cfg_id == config_id:
            edit_index = i
            break
    if edit_index is None:
        cli.err_exit(f"Unable to locate config with id '{config_id}'")
    config.set_last_updated()
    routes[edit_index] = config.route
    notif_pol[lib.ROUTES_FIELD] = routes
    api.put_notification_policy(*ctx.get_api_data(), notif_pol)
    cli.try_log(f"Successfully edited Notification Config '{config.id}'")


def apply_tgt_edits(edit_dict: Dict, target_id: str):
    """
    Apply edits to a notification target identified by target_id.

    Args:
        edit_dict (Dict): A dictionary containing the edited target resource.
        target_id (str): The ID of the target to be edited.

    Raises:
        ValueError: If the target with the specified ID is not found.

    """
    ctx = cfg.get_current_context()
    notif_pol = api.get_notification_policy(*ctx.get_api_data())
    target = nt.Target(target_resource=edit_dict)
    target.id = target_id
    targets: Dict = notif_pol.get(lib.TARGETS_FIELD, {})
    old_tgt = None
    for tgt_name, tgt_data in targets.items():
        tgt_id = tgt_data.get(lib.DATA_FIELD, {}).get(lib.ID_FIELD)
        if not tgt_id:
            continue
        if tgt_id == target.id:
            old_tgt = {tgt_name: tgt_data}
            break
        if tgt_name == target.name:
            cli.err_exit("Target names must be unique!")
    if not old_tgt:
        cli.err_exit(f"Unable to locate config with id '{target_id}'")
    old_name = next(iter(old_tgt))
    targets.pop(old_name)
    target.set_last_update_time()
    new_tgt = target.as_target()
    targets.update(**new_tgt)
    notif_pol[lib.TARGETS_FIELD] = targets
    api.put_notification_policy(*ctx.get_api_data(), notif_pol)
    cli.try_log(f"Successfully edited Notification Target '{target.id}'")


def apply_policy_edits(edit_dict: Dict, policy_id: str):
    """
    Apply edits to a policy identified by policy_id.

    Args:
        edit_dict (Dict): A dictionary containing the edited policy resource.
        policy_id (str): The ID of the policy to be edited.
    """
    ctx = cfg.get_current_context()
    pol_type = edit_dict[lib.METADATA_FIELD][lib.METADATA_TYPE_FIELD]
    # Ensure the user didn't manipulate the uid
    edit_dict[lib.METADATA_FIELD][lib.METADATA_UID_FIELD] = policy_id
    api.put_policy_update(*ctx.get_api_data(), edit_dict)
    cli.try_log(
        f"Successfully edited {__pol_resrc_name(pol_type)} '{policy_id}'"
    )


def apply_ruleset_edits(edit_dict: Dict, ruleset_id: str):
    ctx = cfg.get_current_context()
    edit_dict[lib.METADATA_FIELD][lib.METADATA_UID_FIELD] = ruleset_id
    api.put_ruleset_update(*ctx.get_api_data(), edit_dict)
    cli.try_log(f"Successfully edited Ruleset '{ruleset_id}'")


def apply_file_edits(resource, file: IO):
    """
    Apply edits to a file based on the given resource.

    Args:
        resource: The resource containing the edits.
        file (IO): The file to apply the edits to.

    """
    extension = ".json" if file.name.endswith(".json") else ".yaml"
    try:
        file.close()
        with open(file.name, "w", encoding="UTF-8") as f:
            if extension == ".json":
                f.write(json.dumps(resource, sort_keys=False, indent=2))
            else:
                f.write(cli.make_yaml(resource))
    except Exception as e:
        cli.err_exit(f"Unable to write output to {file.name}", exception=e)
    cli.try_log(f"Successfully edited resource file '{file.name}'")


def __strip_comments(yaml_string: str) -> str:
    lines = []
    for line in yaml_string.split("\n"):
        if line.strip().startswith("#"):
            continue
        lines.append(line)
    return "\n".join(lines)


def __add_edit_prompt(yaml_string: str):
    return EDIT_PROMPT + yaml_string


def __add_error_comments(
    resource: str, yaml_string: str, error: str, name: str = None
):
    if isinstance(name, TextIOWrapper):
        name = f' "{name.name}" '
    else:
        name = f' "{name}" ' if name else ""
    yaml_string = __strip_comments(yaml_string)
    error_prompt = f"# {resource}{name}was not valid:\n"
    error = error.split("\n")
    error = ["# " + line for line in error]
    error = "\n".join(error)
    error_prompt += error + "\n#\n"
    rv = EDIT_PROMPT + error_prompt + yaml_string
    return rv


POL_TYPE_TO_RESOURCE_NAME: Dict[str, str] = {
    lib.POL_TYPE_TRACE: lib.SUPPRESSION_POLICY_RESOURCE.name,
    lib.POL_TYPE_SVC: lib.LINUX_SVC_POL_RESOURCE.name,
    lib.POL_TYPE_CONT: lib.CONTAINER_POL_RESOURCE.name,
}


def __pol_resrc_name(policy_type: str) -> str:
    return POL_TYPE_TO_RESOURCE_NAME.get(
        policy_type, lib.POLICIES_RESOURCE.name
    )
